@@ -6,9 +6,12 @@
 #ifndef LIB_JXL_JPEG_ENC_JPEG_DATA_H_
 #define LIB_JXL_JPEG_ENC_JPEG_DATA_H_
 
+#include <jxl/memory_manager.h>
+
 #include <cstdint>
 #include <vector>
 
+#include "lib/jxl/base/span.h"
 #include "lib/jxl/base/status.h"
 #include "lib/jxl/color_encoding_internal.h"
 #include "lib/jxl/enc_params.h"
@@ -20,7 +23,8 @@ namespace jxl {
 class CodecInOut;
 
 namespace jpeg {
-Status EncodeJPEGData(JPEGData& jpeg_data, std::vector<uint8_t>* bytes,
+Status EncodeJPEGData(JxlMemoryManager* memory_manager, JPEGData& jpeg_data,
+                      std::vector<uint8_t>* bytes,
                       const CompressParams& cparams);
 
 void SetColorEncodingFromJpegData(const jpeg::JPEGData& jpg,
