@@ -5,11 +5,13 @@
 
 #include <jxl/cms.h>
 #include <jxl/encode.h>
+#include <jxl/memory_manager.h>
 #include <jxl/types.h>
 
 #include <cmath>
 #include <cstddef>
 #include <cstdint>
+#include <memory>
 #include <sstream>
 #include <string>
 #include <utility>
@@ -20,6 +22,7 @@
 #include "lib/extras/enc/jxl.h"
 #include "lib/extras/metrics.h"
 #include "lib/extras/packed_image.h"
+#include "lib/jxl/base/common.h"
 #include "lib/jxl/base/compiler_specific.h"
 #include "lib/jxl/base/data_parallel.h"
 #include "lib/jxl/base/random.h"
@@ -27,6 +30,7 @@
 #include "lib/jxl/base/status.h"
 #include "lib/jxl/codec_in_out.h"
 #include "lib/jxl/color_encoding_internal.h"
+#include "lib/jxl/common.h"
 #include "lib/jxl/dec_bit_reader.h"
 #include "lib/jxl/enc_aux_out.h"
 #include "lib/jxl/enc_bit_writer.h"
@@ -48,28 +52,30 @@
 #include "lib/jxl/modular/transform/transform.h"
 #include "lib/jxl/padded_bytes.h"
 #include "lib/jxl/test_image.h"
+#include "lib/jxl/test_memory_manager.h"
 #include "lib/jxl/test_utils.h"
 #include "lib/jxl/testing.h"
 
 namespace jxl {
 namespace {
 
-using test::ButteraugliDistance;
-using test::ReadTestData;
-using test::Roundtrip;
-using test::TestImage;
+using ::jxl::test::ButteraugliDistance;
+using ::jxl::test::ReadTestData;
+using ::jxl::test::Roundtrip;
+using ::jxl::test::TestImage;
 
 void TestLosslessGroups(size_t group_size_shift) {
   const std::vector<uint8_t> orig = ReadTestData("jxl/flower/flower.png");
   TestImage t;
-  t.DecodeFromBytes(orig).ClearMetadata();
-  t.SetDimensions(t.ppf().xsize() / 4, t.ppf().ysize() / 4);
+  ASSERT_TRUE(t.DecodeFromBytes(orig));
+  t.ClearMetadata();
+  ASSERT_TRUE(t.SetDimensions(t.ppf().xsize() / 4, t.ppf().ysize() / 4));
 
   extras::JXLCompressParams cparams;
   cparams.distance = 0.0f;
   cparams.AddOption(JXL_ENC_FRAME_SETTING_MODULAR_GROUP_SIZE, group_size_shift);
   extras::JXLDecompressParams dparams;
-  dparams.accepted_formats = {{3, JXL_TYPE_UINT16, JXL_LITTLE_ENDIAN, 0}};
+  dparams.accepted_formats = {{3, JXL_TYPE_UINT8, JXL_LITTLE_ENDIAN, 0}};
 
   extras::PackedPixelFile ppf_out;
   size_t compressed_size =
@@ -80,20 +86,21 @@ void TestLosslessGroups(size_t group_size_shift) {
 
 TEST(ModularTest, RoundtripLosslessGroups128) { TestLosslessGroups(0); }
 
-TEST(ModularTest, JXL_TSAN_SLOW_TEST(RoundtripLosslessGroups512)) {
+JXL_TSAN_SLOW_TEST(ModularTest, RoundtripLosslessGroups512) {
   TestLosslessGroups(2);
 }
 
-TEST(ModularTest, JXL_TSAN_SLOW_TEST(RoundtripLosslessGroups1024)) {
+JXL_TSAN_SLOW_TEST(ModularTest, RoundtripLosslessGroups1024) {
   TestLosslessGroups(3);
 }
 
-TEST(ModularTest, RoundtripLosslessCustomWP_PermuteRCT) {
+TEST(ModularTest, RoundtripLosslessCustomWpPermuteRCT) {
   const std::vector<uint8_t> orig =
       ReadTestData("external/wesaturate/500px/u76c0g_bliznaca_srgb8.png");
   TestImage t;
-  t.DecodeFromBytes(orig).ClearMetadata();
-  t.SetDimensions(100, 100);
+  ASSERT_TRUE(t.DecodeFromBytes(orig));
+  t.ClearMetadata();
+  ASSERT_TRUE(t.SetDimensions(100, 100));
 
   extras::JXLCompressParams cparams;
   cparams.distance = 0.0f;
@@ -104,7 +111,7 @@ TEST(ModularTest, RoundtripLosslessCustomWP_PermuteRCT) {
   // slowest speed so different WP modes are tried
   cparams.AddOption(JXL_ENC_FRAME_SETTING_EFFORT, 9);
   extras::JXLDecompressParams dparams;
-  dparams.accepted_formats = {{3, JXL_TYPE_UINT16, JXL_LITTLE_ENDIAN, 0}};
+  dparams.accepted_formats = {{3, JXL_TYPE_UINT8, JXL_LITTLE_ENDIAN, 0}};
 
   extras::PackedPixelFile ppf_out;
   size_t compressed_size =
@@ -114,6 +121,7 @@ TEST(ModularTest, RoundtripLosslessCustomWP_PermuteRCT) {
 }
 
 TEST(ModularTest, RoundtripLossyDeltaPalette) {
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
   const std::vector<uint8_t> orig =
       ReadTestData("external/wesaturate/500px/u76c0g_bliznaca_srgb8.png");
   CompressParams cparams;
@@ -122,45 +130,50 @@ TEST(ModularTest, RoundtripLossyDeltaPalette) {
   cparams.lossy_palette = true;
   cparams.palette_colors = 0;
 
-  CodecInOut io_out;
+  CodecInOut io_out{memory_manager};
 
-  CodecInOut io;
+  CodecInOut io{memory_manager};
   ASSERT_TRUE(SetFromBytes(Bytes(orig), &io));
-  io.ShrinkTo(300, 100);
+  ASSERT_TRUE(io.ShrinkTo(300, 100));
 
   size_t compressed_size;
   JXL_EXPECT_OK(Roundtrip(&io, cparams, {}, &io_out, _, &compressed_size));
   EXPECT_LE(compressed_size, 6800u);
-  EXPECT_THAT(ButteraugliDistance(io.frames, io_out.frames, ButteraugliParams(),
-                                  *JxlGetDefaultCms(),
-                                  /*distmap=*/nullptr),
-              IsSlightlyBelow(1.5));
+  EXPECT_SLIGHTLY_BELOW(
+      ButteraugliDistance(io.frames, io_out.frames, ButteraugliParams(),
+                          *JxlGetDefaultCms(),
+                          /*distmap=*/nullptr),
+      1.5);
 }
 TEST(ModularTest, RoundtripLossyDeltaPaletteWP) {
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
   const std::vector<uint8_t> orig =
       ReadTestData("external/wesaturate/500px/u76c0g_bliznaca_srgb8.png");
   CompressParams cparams;
   cparams.SetLossless();
   cparams.lossy_palette = true;
   cparams.palette_colors = 0;
+  // TODO(jon): this is currently ignored, and Avg4 is always used instead
   cparams.options.predictor = jxl::Predictor::Weighted;
 
-  CodecInOut io_out;
+  CodecInOut io_out{memory_manager};
 
-  CodecInOut io;
+  CodecInOut io{memory_manager};
   ASSERT_TRUE(SetFromBytes(Bytes(orig), &io));
-  io.ShrinkTo(300, 100);
+  ASSERT_TRUE(io.ShrinkTo(300, 100));
 
   size_t compressed_size;
   JXL_EXPECT_OK(Roundtrip(&io, cparams, {}, &io_out, _, &compressed_size));
-  EXPECT_LE(compressed_size, 7000u);
-  EXPECT_THAT(ButteraugliDistance(io.frames, io_out.frames, ButteraugliParams(),
-                                  *JxlGetDefaultCms(),
-                                  /*distmap=*/nullptr),
-              IsSlightlyBelow(10.1));
+  EXPECT_LE(compressed_size, 6500u);
+  EXPECT_SLIGHTLY_BELOW(
+      ButteraugliDistance(io.frames, io_out.frames, ButteraugliParams(),
+                          *JxlGetDefaultCms(),
+                          /*distmap=*/nullptr),
+      1.5);
 }
 
 TEST(ModularTest, RoundtripLossy) {
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
   const std::vector<uint8_t> orig =
       ReadTestData("external/wesaturate/500px/u76c0g_bliznaca_srgb8.png");
   CompressParams cparams;
@@ -168,50 +181,55 @@ TEST(ModularTest, RoundtripLossy) {
   cparams.butteraugli_distance = 2.f;
   cparams.SetCms(*JxlGetDefaultCms());
 
-  CodecInOut io_out;
+  CodecInOut io_out{memory_manager};
 
-  CodecInOut io;
+  CodecInOut io{memory_manager};
   ASSERT_TRUE(SetFromBytes(Bytes(orig), &io));
 
   size_t compressed_size;
   JXL_EXPECT_OK(Roundtrip(&io, cparams, {}, &io_out, _, &compressed_size));
   EXPECT_LE(compressed_size, 30000u);
-  EXPECT_THAT(ButteraugliDistance(io.frames, io_out.frames, ButteraugliParams(),
-                                  *JxlGetDefaultCms(),
-                                  /*distmap=*/nullptr),
-              IsSlightlyBelow(2.3));
+  EXPECT_SLIGHTLY_BELOW(
+      ButteraugliDistance(io.frames, io_out.frames, ButteraugliParams(),
+                          *JxlGetDefaultCms(),
+                          /*distmap=*/nullptr),
+      2.3);
 }
 
 TEST(ModularTest, RoundtripLossy16) {
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
   const std::vector<uint8_t> orig =
       ReadTestData("external/raw.pixls/DJI-FC6310-16bit_709_v4_krita.png");
   CompressParams cparams;
   cparams.modular_mode = true;
   cparams.butteraugli_distance = 2.f;
 
-  CodecInOut io_out;
+  CodecInOut io_out{memory_manager};
 
-  CodecInOut io;
+  CodecInOut io{memory_manager};
   ASSERT_TRUE(SetFromBytes(Bytes(orig), &io));
-  JXL_CHECK(!io.metadata.m.have_preview);
-  JXL_CHECK(io.frames.size() == 1);
-  JXL_CHECK(
+  ASSERT_TRUE(!io.metadata.m.have_preview);
+  ASSERT_TRUE(io.frames.size() == 1);
+  ASSERT_TRUE(
       io.frames[0].TransformTo(ColorEncoding::SRGB(), *JxlGetDefaultCms()));
   io.metadata.m.color_encoding = ColorEncoding::SRGB();
 
   size_t compressed_size;
   JXL_EXPECT_OK(Roundtrip(&io, cparams, {}, &io_out, _, &compressed_size));
   EXPECT_LE(compressed_size, 300u);
-  EXPECT_THAT(ButteraugliDistance(io.frames, io_out.frames, ButteraugliParams(),
-                                  *JxlGetDefaultCms(),
-                                  /*distmap=*/nullptr),
-              IsSlightlyBelow(1.6));
+  EXPECT_SLIGHTLY_BELOW(
+      ButteraugliDistance(io.frames, io_out.frames, ButteraugliParams(),
+                          *JxlGetDefaultCms(),
+                          /*distmap=*/nullptr),
+      1.6);
 }
 
 TEST(ModularTest, RoundtripExtraProperties) {
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
   constexpr size_t kSize = 250;
-  JXL_ASSIGN_OR_DIE(Image image,
-                    Image::Create(kSize, kSize, /*bitdepth=*/8, 3));
+  JXL_TEST_ASSIGN_OR_DIE(Image image,
+                         Image::Create(memory_manager, kSize, kSize,
+                                       /*bitdepth=*/8, 3));
   ModularOptions options;
   options.max_properties = 4;
   options.predictor = Predictor::Zero;
@@ -223,20 +241,22 @@ TEST(ModularTest, RoundtripExtraProperties) {
     }
   }
   ZeroFillImage(&image.channel[1].plane);
-  BitWriter writer;
+  BitWriter writer{memory_manager};
   ASSERT_TRUE(ModularGenericCompress(image, options, &writer));
   writer.ZeroPadToByte();
-  JXL_ASSIGN_OR_DIE(Image decoded, Image::Create(kSize, kSize, /*bitdepth=*/8,
-                                                 image.channel.size()));
+  JXL_TEST_ASSIGN_OR_DIE(Image decoded,
+                         Image::Create(memory_manager, kSize, kSize,
+                                       /*bitdepth=*/8, image.channel.size()));
   for (size_t i = 0; i < image.channel.size(); i++) {
     const Channel& ch = image.channel[i];
-    JXL_ASSIGN_OR_DIE(decoded.channel[i],
-                      Channel::Create(ch.w, ch.h, ch.hshift, ch.vshift));
+    JXL_TEST_ASSIGN_OR_DIE(
+        decoded.channel[i],
+        Channel::Create(memory_manager, ch.w, ch.h, ch.hshift, ch.vshift));
   }
   Status status = true;
   {
     BitReader reader(writer.GetSpan());
-    BitReaderScopedCloser closer(&reader, &status);
+    BitReaderScopedCloser closer(reader, status);
     ASSERT_TRUE(ModularGenericDecompress(&reader, decoded, /*header=*/nullptr,
                                          /*group_id=*/0, &options));
   }
@@ -283,6 +303,7 @@ JXL_GTEST_INSTANTIATE_TEST_SUITE_P(RoundtripLossless, ModularTestParam,
                                    LosslessTestDescription);
 
 TEST_P(ModularTestParam, RoundtripLossless) {
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
   RoundtripLosslessConfig config = GetParam();
   int bitdepth = config.bitdepth;
   int responsive = config.responsive;
@@ -291,7 +312,7 @@ TEST_P(ModularTestParam, RoundtripLossless) {
   Rng generator(123);
   const std::vector<uint8_t> orig =
       ReadTestData("external/wesaturate/500px/u76c0g_bliznaca_srgb8.png");
-  CodecInOut io1;
+  CodecInOut io1{memory_manager};
   ASSERT_TRUE(SetFromBytes(Bytes(orig), &io1, pool));
 
   // vary the dimensions a bit, in case of bugs related to
@@ -299,14 +320,15 @@ TEST_P(ModularTestParam, RoundtripLossless) {
   size_t xsize = 423 + bitdepth;
   size_t ysize = 467 + bitdepth;
 
-  CodecInOut io;
-  io.SetSize(xsize, ysize);
+  CodecInOut io{memory_manager};
+  ASSERT_TRUE(io.SetSize(xsize, ysize));
   io.metadata.m.color_encoding = jxl::ColorEncoding::SRGB(false);
   io.metadata.m.SetUintSamples(bitdepth);
 
   double factor = ((1lu << bitdepth) - 1lu);
   double ifactor = 1.0 / factor;
-  JXL_ASSIGN_OR_DIE(Image3F noise_added, Image3F::Create(xsize, ysize));
+  JXL_TEST_ASSIGN_OR_DIE(Image3F noise_added,
+                         Image3F::Create(memory_manager, xsize, ysize));
 
   for (size_t c = 0; c < 3; c++) {
     for (size_t y = 0; y < ysize; y++) {
@@ -322,7 +344,8 @@ TEST_P(ModularTestParam, RoundtripLossless) {
       }
     }
   }
-  io.SetFromImage(std::move(noise_added), jxl::ColorEncoding::SRGB(false));
+  ASSERT_TRUE(
+      io.SetFromImage(std::move(noise_added), jxl::ColorEncoding::SRGB(false)));
 
   CompressParams cparams;
   cparams.modular_mode = true;
@@ -331,10 +354,10 @@ TEST_P(ModularTestParam, RoundtripLossless) {
   cparams.options.predictor = {Predictor::Zero};
   cparams.speed_tier = SpeedTier::kThunder;
   cparams.responsive = responsive;
-  CodecInOut io2;
+  CodecInOut io2{memory_manager};
   size_t compressed_size;
   JXL_EXPECT_OK(Roundtrip(&io, cparams, {}, &io2, _, &compressed_size));
-  EXPECT_LE(compressed_size, bitdepth * xsize * ysize / 3 * 1.1);
+  EXPECT_LE(compressed_size, bitdepth * xsize * ysize / 3.0 * 1.1);
   EXPECT_LE(0, ComputeDistance2(io.Main(), io2.Main(), *JxlGetDefaultCms()));
   size_t different = 0;
   for (size_t c = 0; c < 3; c++) {
@@ -342,8 +365,8 @@ TEST_P(ModularTestParam, RoundtripLossless) {
       const float* in = io.Main().color()->PlaneRow(c, y);
       const float* out = io2.Main().color()->PlaneRow(c, y);
       for (size_t x = 0; x < xsize; x++) {
-        uint32_t uin = in[x] * factor + 0.5;
-        uint32_t uout = out[x] * factor + 0.5;
+        uint32_t uin = std::lroundf(in[x] * factor);
+        uint32_t uout = std::lroundf(out[x] * factor);
         // check that the integer values are identical
         if (uin != uout) different++;
       }
@@ -353,10 +376,11 @@ TEST_P(ModularTestParam, RoundtripLossless) {
 }
 
 TEST(ModularTest, RoundtripLosslessCustomFloat) {
-  CodecInOut io;
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
+  CodecInOut io{memory_manager};
   size_t xsize = 100;
   size_t ysize = 300;
-  io.SetSize(xsize, ysize);
+  ASSERT_TRUE(io.SetSize(xsize, ysize));
   io.metadata.m.bit_depth.bits_per_sample = 18;
   io.metadata.m.bit_depth.exponent_bits_per_sample = 6;
   io.metadata.m.bit_depth.floating_point_sample = true;
@@ -364,7 +388,8 @@ TEST(ModularTest, RoundtripLosslessCustomFloat) {
   ColorEncoding color_encoding;
   color_encoding.Tf().SetTransferFunction(TransferFunction::kLinear);
   color_encoding.SetColorSpace(ColorSpace::kRGB);
-  JXL_ASSIGN_OR_DIE(Image3F testimage, Image3F::Create(xsize, ysize));
+  JXL_TEST_ASSIGN_OR_DIE(Image3F testimage,
+                         Image3F::Create(memory_manager, xsize, ysize));
   float factor = 1.f / (1 << 14);
   for (size_t c = 0; c < 3; c++) {
     for (size_t y = 0; y < ysize; y++) {
@@ -374,7 +399,7 @@ TEST(ModularTest, RoundtripLosslessCustomFloat) {
       }
     }
   }
-  io.SetFromImage(std::move(testimage), color_encoding);
+  ASSERT_TRUE(io.SetFromImage(std::move(testimage), color_encoding));
   io.metadata.m.color_encoding = color_encoding;
   io.metadata.m.SetIntensityTarget(255);
 
@@ -386,7 +411,7 @@ TEST(ModularTest, RoundtripLosslessCustomFloat) {
   cparams.speed_tier = SpeedTier::kThunder;
   cparams.decoding_speed_tier = 2;
 
-  CodecInOut io2;
+  CodecInOut io2{memory_manager};
   size_t compressed_size;
   JXL_EXPECT_OK(Roundtrip(&io, cparams, {}, &io2, _, &compressed_size));
   EXPECT_LE(compressed_size, 23000u);
@@ -394,19 +419,23 @@ TEST(ModularTest, RoundtripLosslessCustomFloat) {
 }
 
 void WriteHeaders(BitWriter* writer, size_t xsize, size_t ysize) {
-  BitWriter::Allotment allotment(writer, 16);
-  writer->Write(8, 0xFF);
-  writer->Write(8, kCodestreamMarker);
-  allotment.ReclaimAndCharge(writer, 0, nullptr);
+  ASSERT_TRUE(writer->WithMaxBits(16, LayerType::Header, nullptr, [&] {
+    writer->Write(8, 0xFF);
+    writer->Write(8, kCodestreamMarker);
+    return true;
+  }));
   CodecMetadata metadata;
   EXPECT_TRUE(metadata.size.Set(xsize, ysize));
-  EXPECT_TRUE(WriteSizeHeader(metadata.size, writer, 0, nullptr));
+  EXPECT_TRUE(
+      WriteSizeHeader(metadata.size, writer, LayerType::Header, nullptr));
   metadata.m.color_encoding = ColorEncoding::LinearSRGB(/*is_gray=*/true);
   metadata.m.xyb_encoded = false;
   metadata.m.SetUintSamples(31);
-  EXPECT_TRUE(WriteImageMetadata(metadata.m, writer, 0, nullptr));
+  EXPECT_TRUE(
+      WriteImageMetadata(metadata.m, writer, LayerType::Header, nullptr));
   metadata.transform_data.nonserialized_xyb_encoded = metadata.m.xyb_encoded;
-  EXPECT_TRUE(Bundle::Write(metadata.transform_data, writer, 0, nullptr));
+  EXPECT_TRUE(Bundle::Write(metadata.transform_data, writer, LayerType::Header,
+                            nullptr));
   writer->ZeroPadToByte();
   FrameHeader frame_header(&metadata);
   frame_header.encoding = FrameEncoding::kModular;
@@ -441,27 +470,30 @@ void WriteHistograms(BitWriter* writer) {
 }
 
 TEST(ModularTest, PredictorIntegerOverflow) {
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
   const size_t xsize = 1;
   const size_t ysize = 1;
-  BitWriter writer;
+  BitWriter writer{memory_manager};
   WriteHeaders(&writer, xsize, ysize);
-  std::vector<BitWriter> group_codes(1);
+  std::vector<std::unique_ptr<BitWriter>> group_codes;
+  group_codes.emplace_back(jxl::make_unique<BitWriter>(memory_manager));
   {
-    BitWriter* bw = group_codes.data();
-    BitWriter::Allotment allotment(bw, 1 << 20);
-    WriteHistograms(bw);
-    GroupHeader header;
-    header.use_global_tree = true;
-    EXPECT_TRUE(Bundle::Write(header, bw, 0, nullptr));
-    // After UnpackSigned this becomes (1 << 31) - 1, the largest pixel_type,
-    // and after adding the offset we get -(1 << 31).
-    bw->Write(8, 119);
-    bw->Write(28, 0xfffffff);
-    bw->ZeroPadToByte();
-    allotment.ReclaimAndCharge(bw, 0, nullptr);
+    std::unique_ptr<BitWriter>& bw = group_codes[0];
+    ASSERT_TRUE(bw->WithMaxBits(1 << 20, LayerType::Header, nullptr, [&] {
+      WriteHistograms(bw.get());
+      GroupHeader header;
+      header.use_global_tree = true;
+      EXPECT_TRUE(Bundle::Write(header, bw.get(), LayerType::Header, nullptr));
+      // After UnpackSigned this becomes (1 << 31) - 1, the largest pixel_type,
+      // and after adding the offset we get -(1 << 31).
+      bw->Write(8, 119);
+      bw->Write(28, 0xfffffff);
+      bw->ZeroPadToByte();
+      return true;
+    }));
   }
   EXPECT_TRUE(WriteGroupOffsets(group_codes, {}, &writer, nullptr));
-  writer.AppendByteAligned(group_codes);
+  ASSERT_TRUE(writer.AppendByteAligned(group_codes));
 
   PaddedBytes compressed = std::move(writer).TakeBytes();
   extras::PackedPixelFile ppf;
@@ -476,40 +508,43 @@ TEST(ModularTest, PredictorIntegerOverflow) {
 }
 
 TEST(ModularTest, UnsqueezeIntegerOverflow) {
+  JxlMemoryManager* memory_manager = jxl::test::MemoryManager();
   // Image width is 9 so we can test both the SIMD and non-vector code paths.
   const size_t xsize = 9;
   const size_t ysize = 2;
-  BitWriter writer;
+  BitWriter writer{memory_manager};
   WriteHeaders(&writer, xsize, ysize);
-  std::vector<BitWriter> group_codes(1);
+  std::vector<std::unique_ptr<BitWriter>> group_codes;
+  group_codes.emplace_back(jxl::make_unique<BitWriter>(memory_manager));
   {
-    BitWriter* bw = group_codes.data();
-    BitWriter::Allotment allotment(bw, 1 << 20);
-    WriteHistograms(bw);
-    GroupHeader header;
-    header.use_global_tree = true;
-    header.transforms.emplace_back();
-    header.transforms[0].id = TransformId::kSqueeze;
-    SqueezeParams params;
-    params.horizontal = false;
-    params.in_place = true;
-    params.begin_c = 0;
-    params.num_c = 1;
-    header.transforms[0].squeezes.emplace_back(params);
-    EXPECT_TRUE(Bundle::Write(header, bw, 0, nullptr));
-    for (size_t i = 0; i < xsize * ysize; ++i) {
-      // After UnpackSigned and adding offset, this becomes (1 << 31) - 1, both
-      // in the image and in the residual channels, and unsqueeze makes them
-      // ~(3 << 30) and (1 << 30) (in pixel_type_w) and the first wraps around
-      // to about -(1 << 30).
-      bw->Write(8, 119);
-      bw->Write(28, 0xffffffe);
-    }
-    bw->ZeroPadToByte();
-    allotment.ReclaimAndCharge(bw, 0, nullptr);
+    std::unique_ptr<BitWriter>& bw = group_codes[0];
+    ASSERT_TRUE(bw->WithMaxBits(1 << 20, LayerType::Header, nullptr, [&] {
+      WriteHistograms(bw.get());
+      GroupHeader header;
+      header.use_global_tree = true;
+      header.transforms.emplace_back();
+      header.transforms[0].id = TransformId::kSqueeze;
+      SqueezeParams params;
+      params.horizontal = false;
+      params.in_place = true;
+      params.begin_c = 0;
+      params.num_c = 1;
+      header.transforms[0].squeezes.emplace_back(params);
+      EXPECT_TRUE(Bundle::Write(header, bw.get(), LayerType::Header, nullptr));
+      for (size_t i = 0; i < xsize * ysize; ++i) {
+        // After UnpackSigned and adding offset, this becomes (1 << 31) - 1,
+        // both in the image and in the residual channels, and unsqueeze makes
+        // them ~(3 << 30) and (1 << 30) (in pixel_type_w) and the first wraps
+        // around to about -(1 << 30).
+        bw->Write(8, 119);
+        bw->Write(28, 0xffffffe);
+      }
+      bw->ZeroPadToByte();
+      return true;
+    }));
   }
   EXPECT_TRUE(WriteGroupOffsets(group_codes, {}, &writer, nullptr));
-  writer.AppendByteAligned(group_codes);
+  ASSERT_TRUE(writer.AppendByteAligned(group_codes));
 
   PaddedBytes compressed = std::move(writer).TakeBytes();
   extras::PackedPixelFile ppf;
