@@ -4,7 +4,9 @@
 // license that can be found in the LICENSE file.
 
 #include <jxl/cms.h>
+#include <jxl/cms_interface.h>
 #include <jxl/decode.h>
+#include <jxl/memory_manager.h>
 #include <jxl/types.h>
 
 #include <algorithm>
@@ -28,6 +30,7 @@
 #include "lib/extras/metrics.h"
 #include "lib/extras/packed_image.h"
 #include "lib/extras/packed_image_convert.h"
+#include "lib/jxl/base/common.h"
 #include "lib/jxl/base/compiler_specific.h"
 #include "lib/jxl/base/data_parallel.h"
 #include "lib/jxl/base/printf_macros.h"
@@ -35,7 +38,6 @@
 #include "lib/jxl/base/span.h"
 #include "lib/jxl/base/status.h"
 #include "lib/jxl/butteraugli/butteraugli.h"
-#include "lib/jxl/cache_aligned.h"
 #include "lib/jxl/codec_in_out.h"
 #include "lib/jxl/color_encoding_internal.h"
 #include "lib/jxl/enc_butteraugli_comparator.h"
@@ -48,16 +50,21 @@
 #include "tools/benchmark/benchmark_file_io.h"
 #include "tools/benchmark/benchmark_stats.h"
 #include "tools/benchmark/benchmark_utils.h"
+#include "tools/cmdline.h"
 #include "tools/codec_config.h"
 #include "tools/file_io.h"
+#include "tools/no_memory_manager.h"
 #include "tools/speed_stats.h"
 #include "tools/ssimulacra2.h"
 #include "tools/thread_pool_internal.h"
+#include "tools/tracking_memory_manager.h"
 
 namespace jpegxl {
 namespace tools {
 namespace {
 
+#define QUIT(M) JPEGXL_TOOLS_ABORT(M)
+
 using ::jxl::ButteraugliParams;
 using ::jxl::Bytes;
 using ::jxl::CodecInOut;
@@ -68,26 +75,39 @@ using ::jxl::ImageF;
 using ::jxl::JxlButteraugliComparator;
 using ::jxl::Rng;
 using ::jxl::Status;
+using ::jxl::StatusOr;
 using ::jxl::ThreadPool;
 using ::jxl::extras::PackedPixelFile;
 
 Status WriteImage(const Image3F& image, ThreadPool* pool,
                   const std::string& filename) {
   JxlPixelFormat format = {3, JXL_TYPE_UINT8, JXL_BIG_ENDIAN, 0};
-  PackedPixelFile ppf = jxl::extras::ConvertImage3FToPackedPixelFile(
-      image, ColorEncoding::SRGB(), format, pool);
+  JXL_ASSIGN_OR_RETURN(PackedPixelFile ppf,
+                       jxl::extras::ConvertImage3FToPackedPixelFile(
+                           image, ColorEncoding::SRGB(), format, pool));
   std::vector<uint8_t> encoded;
-  return Encode(ppf, filename, &encoded, pool) && WriteFile(filename, encoded);
+  return jxl::Encode(ppf, filename, &encoded, pool) &&
+         WriteFile(filename, encoded);
+}
+
+void PrintStats(const TrackingMemoryManager& memory_manager) {
+  fprintf(stderr,
+          "Allocation count: %" PRIuS ", total: %E (max bytes in use: %E)\n",
+          static_cast<size_t>(memory_manager.total_allocations),
+          static_cast<double>(memory_manager.total_bytes_allocated),
+          static_cast<double>(memory_manager.max_bytes_in_use));
 }
 
 Status ReadPNG(const std::string& filename, Image3F* image) {
-  CodecInOut io;
+  JxlMemoryManager* memory_manager = jpegxl::tools::NoMemoryManager();
+  CodecInOut io{memory_manager};
   std::vector<uint8_t> encoded;
-  JXL_CHECK(ReadFile(filename, &encoded));
-  JXL_CHECK(
-      jxl::SetFromBytes(jxl::Bytes(encoded), jxl::extras::ColorHints(), &io));
-  JXL_ASSIGN_OR_DIE(*image, Image3F::Create(io.xsize(), io.ysize()));
-  CopyImageTo(*io.Main().color(), image);
+  JXL_RETURN_IF_ERROR(ReadFile(filename, &encoded));
+  JXL_RETURN_IF_ERROR(
+      jxl::SetFromBytes(Bytes(encoded), jxl::extras::ColorHints(), &io));
+  JXL_ASSIGN_OR_RETURN(*image,
+                       Image3F::Create(memory_manager, io.xsize(), io.ysize()));
+  JXL_RETURN_IF_ERROR(CopyImageTo(*io.Main().color(), image));
   return true;
 }
 
@@ -117,19 +137,19 @@ std::string CodecToExtension(const std::string& codec_name, char sep) {
   return result;
 }
 
-void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
-                const std::vector<std::string>& extra_metrics_commands,
-                ImageCodec* codec, ThreadPool* inner_pool,
-                std::vector<uint8_t>* compressed, BenchmarkStats* s) {
+Status DoCompress(const std::string& filename, const PackedPixelFile& ppf,
+                  const std::vector<std::string>& extra_metrics_commands,
+                  ImageCodec* codec, ThreadPool* inner_pool,
+                  std::vector<uint8_t>* compressed, BenchmarkStats* s) {
+  JxlMemoryManager* memory_manager = jpegxl::tools::NoMemoryManager();
   ++s->total_input_files;
 
   if (ppf.frames.size() != 1) {
     // Multiple frames not supported.
-    s->total_errors++;
     if (!Args()->silent_errors) {
       JXL_WARNING("multiframe input image not supported %s", filename.c_str());
     }
-    return;
+    return false;
   }
   const size_t xsize = ppf.info.xsize;
   const size_t ysize = ppf.info.ysize;
@@ -146,64 +166,65 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
     // this function to indicate it as error in the stats.
     valid = false;
   }
+  const PackedPixelFile* ppf1 = &ppf;
+  PackedPixelFile ppf2;
 
-  std::string ext = FileExtension(filename);
-  if (valid && !Args()->decode_only) {
-    for (size_t i = 0; i < Args()->encode_reps; ++i) {
-      if (codec->CanRecompressJpeg() && (ext == ".jpg" || ext == ".jpeg")) {
-        std::vector<uint8_t> data_in;
-        JXL_CHECK(ReadFile(filename, &data_in));
-        JXL_CHECK(
-            codec->RecompressJpeg(filename, data_in, compressed, &speed_stats));
-      } else {
-        Status status = codec->Compress(filename, ppf, inner_pool, compressed,
-                                        &speed_stats);
-        if (!status) {
-          valid = false;
-          if (!Args()->silent_errors) {
-            std::string message = codec->GetErrorMessage();
-            if (!message.empty()) {
-              fprintf(stderr, "Error in %s codec: %s\n",
-                      codec->description().c_str(), message.c_str());
-            } else {
-              fprintf(stderr, "Error in %s codec\n",
-                      codec->description().c_str());
+  for (size_t generation = 0; generation <= Args()->generations; generation++) {
+    std::string ext = FileExtension(filename);
+    if (valid && !Args()->decode_only) {
+      for (size_t i = 0; i < Args()->encode_reps; ++i) {
+        if (codec->CanRecompressJpeg() && (ext == ".jpg" || ext == ".jpeg")) {
+          std::vector<uint8_t> data_in;
+          JXL_RETURN_IF_ERROR(ReadFile(filename, &data_in));
+          JXL_RETURN_IF_ERROR(codec->RecompressJpeg(filename, data_in,
+                                                    compressed, &speed_stats));
+        } else {
+          Status status = codec->Compress(filename, *ppf1, inner_pool,
+                                          compressed, &speed_stats);
+          if (!status) {
+            valid = false;
+            if (!Args()->silent_errors) {
+              std::string message = codec->GetErrorMessage();
+              if (!message.empty()) {
+                fprintf(stderr, "Error in %s codec: %s\n",
+                        codec->description().c_str(), message.c_str());
+              } else {
+                fprintf(stderr, "Error in %s codec\n",
+                        codec->description().c_str());
+              }
             }
           }
         }
       }
+      JXL_RETURN_IF_ERROR(speed_stats.GetSummary(&summary));
+      s->total_time_encode += summary.central_tendency;
     }
-    JXL_CHECK(speed_stats.GetSummary(&summary));
-    s->total_time_encode += summary.central_tendency;
-  }
 
-  if (valid && Args()->decode_only) {
-    std::vector<uint8_t> data_in;
-    JXL_CHECK(ReadFile(filename, &data_in));
-    compressed->insert(compressed->end(), data_in.begin(), data_in.end());
-  }
+    if (valid && Args()->decode_only) {
+      std::vector<uint8_t> data_in;
+      JXL_RETURN_IF_ERROR(ReadFile(filename, &data_in));
+      compressed->insert(compressed->end(), data_in.begin(), data_in.end());
+    }
 
-  // Decompress
-  PackedPixelFile ppf2;
-  if (valid) {
-    speed_stats = jpegxl::tools::SpeedStats();
-    for (size_t i = 0; i < Args()->decode_reps; ++i) {
-      if (!codec->Decompress(filename, Bytes(*compressed), inner_pool, &ppf2,
-                             &speed_stats)) {
-        if (!Args()->silent_errors) {
-          fprintf(stderr,
-                  "%s failed to decompress encoded image. Original source:"
-                  " %s\n",
-                  codec->description().c_str(), filename.c_str());
+    // Decompress
+    if (valid) {
+      speed_stats = jpegxl::tools::SpeedStats();
+      for (size_t i = 0; i < Args()->decode_reps; ++i) {
+        if (!codec->Decompress(filename, Bytes(*compressed), inner_pool, &ppf2,
+                               &speed_stats)) {
+          if (!Args()->silent_errors) {
+            fprintf(stderr,
+                    "%s failed to decompress encoded image. Original source:"
+                    " %s\n",
+                    codec->description().c_str(), filename.c_str());
+          }
+          valid = false;
         }
-        valid = false;
       }
+      JXL_RETURN_IF_ERROR(speed_stats.GetSummary(&summary));
+      s->total_time_decode += summary.central_tendency;
     }
-    for (const auto& frame : ppf2.frames) {
-      s->total_input_pixels += frame.color.xsize * frame.color.ysize;
-    }
-    JXL_CHECK(speed_stats.GetSummary(&summary));
-    s->total_time_decode += summary.central_tendency;
+    ppf1 = &ppf2;
   }
 
   std::string name = FileBaseName(filename);
@@ -213,6 +234,12 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
     s->total_errors++;
   }
 
+  if (valid) {
+    for (const auto& frame : ppf2.frames) {
+      s->total_input_pixels += frame.color.xsize * frame.color.ysize;
+    }
+  }
+
   if (ppf.frames.size() != ppf2.frames.size()) {
     if (!Args()->silent_errors) {
       // Animated gifs not supported yet?
@@ -230,10 +257,12 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
   float distance = 1.0f;
 
   if (valid && !skip_butteraugli) {
-    CodecInOut ppf_io;
-    JXL_CHECK(ConvertPackedPixelFileToCodecInOut(ppf, inner_pool, &ppf_io));
-    CodecInOut ppf2_io;
-    JXL_CHECK(ConvertPackedPixelFileToCodecInOut(ppf2, inner_pool, &ppf2_io));
+    CodecInOut ppf_io{memory_manager};
+    JXL_RETURN_IF_ERROR(
+        ConvertPackedPixelFileToCodecInOut(ppf, inner_pool, &ppf_io));
+    CodecInOut ppf2_io{memory_manager};
+    JXL_RETURN_IF_ERROR(
+        ConvertPackedPixelFileToCodecInOut(ppf2, inner_pool, &ppf2_io));
     const ImageBundle& ib1 = ppf_io.Main();
     const ImageBundle& ib2 = ppf2_io.Main();
     if (jxl::SameSize(ppf, ppf2)) {
@@ -246,12 +275,13 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
 
       const JxlCmsInterface& cms = *JxlGetDefaultCms();
       JxlButteraugliComparator comparator(params, cms);
-      JXL_CHECK(ComputeScore(ib1, ib2, &comparator, cms, &distance, &distmap,
-                             inner_pool, codec->IgnoreAlpha()));
+      JXL_RETURN_IF_ERROR(ComputeScore(ib1, ib2, &comparator, cms, &distance,
+                                       &distmap, inner_pool,
+                                       codec->IgnoreAlpha()));
     } else {
       // TODO(veluca): re-upsample and compute proper distance.
       distance = 1e+4f;
-      JXL_ASSIGN_OR_DIE(distmap, ImageF::Create(1, 1));
+      JXL_ASSIGN_OR_RETURN(distmap, ImageF::Create(memory_manager, 1, 1));
       distmap.Row(0)[0] = distance;
     }
     // Update stats
@@ -259,13 +289,16 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
         compressed->empty()
             ? 0
             : jxl::ComputePSNR(ib1, ib2, *JxlGetDefaultCms()) * input_pixels;
-    s->distance_p_norm +=
-        ComputeDistanceP(distmap, ButteraugliParams(), Args()->error_pnorm) *
-        input_pixels;
-    JXL_ASSIGN_OR_DIE(Msssim msssim, ComputeSSIMULACRA2(ib1, ib2));
-    s->ssimulacra2 += msssim.Score() * input_pixels;
+    double pnorm =
+        ComputeDistanceP(distmap, ButteraugliParams(), Args()->error_pnorm);
+    s->distance_p_norm += pnorm * input_pixels;
+    JXL_ASSIGN_OR_RETURN(Msssim msssim, ComputeSSIMULACRA2(ib1, ib2));
+    double ssimulacra2 = msssim.Score();
+    s->ssimulacra2 += ssimulacra2 * input_pixels;
     s->max_distance = std::max(s->max_distance, distance);
     s->distances.push_back(distance);
+    s->pnorms.push_back(pnorm);
+    s->ssimulacra2s.push_back(ssimulacra2);
   }
 
   s->total_compressed_size += compressed->size();
@@ -285,15 +318,15 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
     } else {
       heatmap_fn = compressed_fn + ".heatmap.ppm";
     }
-    JXL_CHECK(MakeDir(outdir));
+    JXL_RETURN_IF_ERROR(MakeDir(outdir));
     if (Args()->save_compressed) {
-      JXL_CHECK(WriteFile(compressed_fn, *compressed));
+      JXL_RETURN_IF_ERROR(WriteFile(compressed_fn, *compressed));
     }
     if (Args()->save_decompressed && valid) {
       // TODO(szabadka): Handle Args()->mul_output
       std::vector<uint8_t> encoded;
-      JXL_CHECK(jxl::Encode(ppf2, decompressed_fn, &encoded));
-      JXL_CHECK(WriteFile(decompressed_fn, encoded));
+      JXL_RETURN_IF_ERROR(jxl::Encode(ppf2, decompressed_fn, &encoded));
+      JXL_RETURN_IF_ERROR(WriteFile(decompressed_fn, encoded));
       if (!skip_butteraugli) {
         float good = Args()->heatmap_good > 0.0f
                          ? Args()->heatmap_good
@@ -302,9 +335,9 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
                         ? Args()->heatmap_bad
                         : jxl::ButteraugliFuzzyInverse(0.5);
         if (Args()->save_heatmap) {
-          JXL_ASSIGN_OR_DIE(Image3F heatmap,
-                            CreateHeatMapImage(distmap, good, bad));
-          JXL_CHECK(WriteImage(heatmap, inner_pool, heatmap_fn));
+          JXL_ASSIGN_OR_RETURN(Image3F heatmap,
+                               CreateHeatMapImage(distmap, good, bad));
+          JXL_RETURN_IF_ERROR(WriteImage(heatmap, inner_pool, heatmap_fn));
         }
       }
     }
@@ -316,21 +349,21 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
     std::string tmp_in_fn;
     std::string tmp_out_fn;
     std::string tmp_res_fn;
-    JXL_CHECK(tmp_in.GetFileName(&tmp_in_fn));
-    JXL_CHECK(tmp_out.GetFileName(&tmp_out_fn));
-    JXL_CHECK(tmp_res.GetFileName(&tmp_res_fn));
+    JXL_RETURN_IF_ERROR(tmp_in.GetFileName(&tmp_in_fn));
+    JXL_RETURN_IF_ERROR(tmp_out.GetFileName(&tmp_out_fn));
+    JXL_RETURN_IF_ERROR(tmp_res.GetFileName(&tmp_res_fn));
 
     std::vector<uint8_t> encoded;
-    JXL_CHECK(jxl::Encode(ppf, tmp_in_fn, &encoded));
-    JXL_CHECK(WriteFile(tmp_in_fn, encoded));
-    JXL_CHECK(jxl::Encode(ppf2, tmp_out_fn, &encoded));
-    JXL_CHECK(WriteFile(tmp_out_fn, encoded));
+    JXL_RETURN_IF_ERROR(jxl::Encode(ppf, tmp_in_fn, &encoded));
+    JXL_RETURN_IF_ERROR(WriteFile(tmp_in_fn, encoded));
+    JXL_RETURN_IF_ERROR(jxl::Encode(ppf2, tmp_out_fn, &encoded));
+    JXL_RETURN_IF_ERROR(WriteFile(tmp_out_fn, encoded));
     // TODO(szabadka) Handle custom intensity target.
     std::string intensity_target = "255";
-    for (size_t i = 0; i < extra_metrics_commands.size(); i++) {
+    for (const auto& extra_metrics_command : extra_metrics_commands) {
       float res = nanf("");
       bool error = false;
-      if (RunCommand(extra_metrics_commands[i],
+      if (RunCommand(extra_metrics_command,
                      {tmp_in_fn, tmp_out_fn, tmp_res_fn, intensity_target})) {
         FILE* f = fopen(tmp_res_fn.c_str(), "r");
         if (fscanf(f, "%f", &res) != 1) {
@@ -343,7 +376,7 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
       if (error) {
         fprintf(stderr,
                 "WARNING: Computation of metric with command %s failed\n",
-                extra_metrics_commands[i].c_str());
+                extra_metrics_command.c_str());
       }
       s->extra_metrics.push_back(res);
     }
@@ -353,6 +386,7 @@ void DoCompress(const std::string& filename, const PackedPixelFile& ppf,
     fprintf(stderr, ".");
     fflush(stderr);
   }
+  return true;
 }
 
 // Makes a base64 data URI for embedded image in HTML
@@ -388,11 +422,11 @@ struct Task {
   BenchmarkStats stats;
 };
 
-void WriteHtmlReport(const std::string& codec_desc,
-                     const std::vector<std::string>& fnames,
-                     const std::vector<const Task*>& tasks,
-                     const std::vector<const PackedPixelFile*>& images,
-                     bool add_heatmap, bool self_contained) {
+Status WriteHtmlReport(const std::string& codec_desc,
+                       const std::vector<std::string>& fnames,
+                       const std::vector<const Task*>& tasks,
+                       const std::vector<const PackedPixelFile*>& images,
+                       bool add_heatmap, bool self_contained) {
   std::string toggle_js =
       "<script type=\"text/javascript\">\n"
       "  var codecname = '" +
@@ -491,8 +525,8 @@ void WriteHtmlReport(const std::string& codec_desc,
   }
   std::string out_html;
   std::string outdir;
-  out_html += "<body bgcolor=\"#000\">\n";
-  out_html += "<style>img { image-rendering: pixelated; }</style>\n";
+  out_html.append("<body bgcolor=\"#000\">\n");
+  out_html.append("<style>img { image-rendering: pixelated; }</style>\n");
   std::string codec_name = codec_desc;
   // Make compatible for filename
   std::replace(codec_name.begin(), codec_name.end(), ':', '_');
@@ -508,8 +542,9 @@ void WriteHtmlReport(const std::string& codec_desc,
         name + CodecToExtension(codec_name, '_') + ".heatmap.png";
 
     const std::string& fname_orig = fnames[i];
-    std::string fname_out = outdir + "/" + name_out;
-    std::string fname_heatmap = outdir + "/" + heatmap_out;
+    std::string fname_out = std::string(outdir).append("/").append(name_out);
+    std::string fname_heatmap =
+        std::string(outdir).append("/").append(heatmap_out);
     std::string url_orig = Args()->originals_url.empty()
                                ? ("file://" + fnames[i])
                                : (Args()->originals_url + "/" + name);
@@ -533,24 +568,41 @@ void WriteHtmlReport(const std::string& codec_desc,
     double max_dist = tasks[i]->stats.max_distance;
     std::string compressed_title = StringPrintf(
         "compressed. bpp: %f, pnorm: %f, max dist: %f", bpp, pnorm, max_dist);
-    out_html += "<div onclick=\"toggle(" + number +
-                ");\" style=\"display:inline-block;width:" + html_width +
-                ";height:" + html_height +
-                ";\">\n"
-                "  <img title=\"" +
-                compressed_title + "\" id=\"preview" + number + "\" src=";
-    out_html += "\"" + url_out + "\"style=\"display:block;\"/>\n";
-    out_html += "  <img title=\"original\" id=\"orig" + number + "\" src=";
-    out_html += "\"" + url_orig + "\"style=\"display:none;\"/>\n";
+    out_html.append("<div onclick=\"toggle(")
+        .append(number)
+        .append(");\" style=\"display:inline-block;width:")
+        .append(html_width)
+        .append(";height:")
+        .append(html_height)
+        .append(";\">\n  <img title=\"")
+        .append(compressed_title)
+        .append("\" id=\"preview")
+        .append(number)
+        .append("\" src=")
+        .append("\"")
+        .append(url_out)
+        .append("\"style=\"display:block;\"/>\n")
+        .append(R"(  <img title="original" id="orig)")
+        .append(number)
+        .append("\" src=")
+        .append("\"")
+        .append(url_orig)
+        .append("\"style=\"display:none;\"/>\n");
     if (add_heatmap) {
-      out_html = "  <img title=\"heatmap\" id=\"hm" + number + "\" src=";
-      out_html += "\"" + url_heatmap + "\"style=\"display:none;\"/>\n";
+      out_html.append(R"(  <img title="heatmap" id="hm)")
+          .append(number)
+          .append("\" src=")
+          .append("\"")
+          .append(url_heatmap)
+          .append("\"style=\"display:none;\"/>\n");
     }
-    out_html += "</div>\n";
+    out_html.append("</div>\n");
   }
-  out_html += "</body>\n";
-  out_html += toggle_js;
-  JXL_CHECK(WriteFile(outdir + "/index." + codec_name + ".html", out_html));
+  out_html.append("</body>\n").append(toggle_js);
+  std::string fname_index =
+      std::string(outdir).append("/index.").append(codec_name).append(".html");
+  JXL_RETURN_IF_ERROR(WriteFile(fname_index, out_html));
+  return true;
 }
 
 // Prints the detailed and aggregate statistics, in the correct order but as
@@ -580,7 +632,7 @@ struct StatPrinter {
     }
   }
 
-  void TaskDone(size_t task_index, const Task& t) {
+  Status TaskDone(size_t task_index, const Task& t) {
     std::lock_guard<std::mutex> guard(mutex);
     tasks_done_++;
     if (Args()->print_details || Args()->show_progress) {
@@ -600,15 +652,15 @@ struct StatPrinter {
       // rendered at the very end, else the details or progress would be
       // rendered in-between the table rows.
       if (tasks_done_ == tasks_->size()) {
-        PrintStatsHeader();
+        JXL_RETURN_IF_ERROR(PrintStatsHeader());
         for (size_t i = 0; i < methods_->size(); i++) {
-          PrintStats((*methods_)[i], i);
+          JXL_RETURN_IF_ERROR(PrintStats((*methods_)[i], i));
         }
-        PrintStatsFooter();
+        JXL_RETURN_IF_ERROR(PrintStatsFooter());
       }
     } else {
       if (tasks_done_ == 1) {
-        PrintStatsHeader();
+        JXL_RETURN_IF_ERROR(PrintStatsHeader());
       }
       // Render lines of the table as soon as it is ready and all previous
       // lines have been printed.
@@ -617,14 +669,16 @@ struct StatPrinter {
           t.idx_method == stats_printed_) {
         while (stats_printed_ < stats_done_.size() &&
                stats_done_[stats_printed_] == fnames_->size()) {
-          PrintStats((*methods_)[stats_printed_], stats_printed_);
+          JXL_RETURN_IF_ERROR(
+              PrintStats((*methods_)[stats_printed_], stats_printed_));
           stats_printed_++;
         }
       }
       if (tasks_done_ == tasks_->size()) {
-        PrintStatsFooter();
+        JXL_RETURN_IF_ERROR(PrintStatsFooter());
       }
     }
+    return true;
   }
 
   void PrintDetails(const Task& t) const {
@@ -646,12 +700,12 @@ struct StatPrinter {
         t.stats.total_input_pixels / (1000000.0 * t.stats.total_time_decode);
     if (Args()->print_details_csv) {
       printf("%s,%s,%" PRIdS ",%" PRIdS ",%" PRIdS
-             ",%.8f,%.8f,%.8f,%.8f,%.8f,%.8f,%.8f,%.8f",
+             ",%.8f,%.8f,%.8f,%.8f,%.8f,%.8f,%.8f,%.8f,%.8f",
              (*methods_)[t.idx_method].c_str(),
              FileBaseName((*fnames_)[t.idx_image]).c_str(),
              t.stats.total_errors, t.stats.total_compressed_size, pixels,
-             enc_mps, dec_mps, comp_bpp, t.stats.max_distance, psnr, p_norm,
-             bpp_p_norm, adj_comp_bpp);
+             enc_mps, dec_mps, comp_bpp, t.stats.max_distance, ssimulacra2,
+             psnr, p_norm, bpp_p_norm, adj_comp_bpp);
       for (float m : t.stats.extra_metrics) {
         printf(",%.8f", m);
       }
@@ -684,7 +738,7 @@ struct StatPrinter {
     fflush(stdout);
   }
 
-  void PrintStats(const std::string& method, size_t idx_method) {
+  Status PrintStats(const std::string& method, size_t idx_method) {
     // Assimilate all tasks with the same idx_method.
     BenchmarkStats method_stats;
     std::vector<const PackedPixelFile*> images;
@@ -696,26 +750,28 @@ struct StatPrinter {
         tasks.push_back(&t);
       }
     }
+    JXL_ENSURE(method_stats.total_input_files == fnames_->size());
 
     std::string out;
 
-    method_stats.PrintMoreStats();  // not concurrent
-    out += method_stats.PrintLine(method, fnames_->size());
+    JXL_RETURN_IF_ERROR(method_stats.PrintMoreStats());  // not concurrent
+    out += method_stats.PrintLine(method);
 
     if (Args()->write_html_report) {
-      WriteHtmlReport(method, *fnames_, tasks, images,
-                      Args()->save_heatmap && Args()->html_report_add_heatmap,
-                      Args()->html_report_self_contained);
+      JXL_RETURN_IF_ERROR(WriteHtmlReport(
+          method, *fnames_, tasks, images,
+          Args()->save_heatmap && Args()->html_report_add_heatmap,
+          Args()->html_report_self_contained));
     }
 
-    stats_aggregate_.push_back(
-        method_stats.ComputeColumns(method, fnames_->size()));
+    stats_aggregate_.push_back(method_stats.ComputeColumns(method));
 
     printf("%s", out.c_str());
     fflush(stdout);
+    return true;
   }
 
-  void PrintStatsHeader() const {
+  Status PrintStatsHeader() const {
     if (Args()->markdown) {
       if (Args()->show_progress) {
         fprintf(stderr, "\n");
@@ -723,18 +779,27 @@ struct StatPrinter {
       }
       printf("```\n");
     }
-    if (fnames_->size() == 1) printf("%s\n", (*fnames_)[0].c_str());
-    printf("%s", PrintHeader(*extra_metrics_names_).c_str());
+    if (fnames_->size() == 1) {
+      printf("%s\n", (*fnames_)[0].c_str());
+    } else {
+      printf("%" PRIuS " images\n", fnames_->size());
+    }
+    JXL_ASSIGN_OR_RETURN(std::string header,
+                         PrintHeader(*extra_metrics_names_));
+    printf("%s", header.c_str());
     fflush(stdout);
+    return true;
   }
 
-  void PrintStatsFooter() const {
-    printf(
-        "%s",
-        PrintAggregate(extra_metrics_names_->size(), stats_aggregate_).c_str());
+  Status PrintStatsFooter() const {
+    JXL_ASSIGN_OR_RETURN(
+        std::string aggregate,
+        PrintAggregate(extra_metrics_names_->size(), stats_aggregate_));
+    printf("%s", aggregate.c_str());
     if (Args()->markdown) printf("```\n");
     printf("\n");
     fflush(stdout);
+    return true;
   }
 
   const std::vector<std::string>* methods_;
@@ -763,33 +828,42 @@ class Benchmark {
 
  public:
   // Return the exit code of the program.
-  static int Run() {
-    int ret = EXIT_SUCCESS;
+  static Status Run() {
+    TrackingMemoryManager memory_manager{};
+    bool ok = true;
     {
       const StringVec methods = GetMethods();
       const StringVec extra_metrics_names = GetExtraMetricsNames();
       const StringVec extra_metrics_commands = GetExtraMetricsCommands();
       const StringVec fnames = GetFilenames();
       // (non-const because Task.stats are updated)
-      std::vector<Task> tasks = CreateTasks(methods, fnames);
+      JXL_ASSIGN_OR_RETURN(std::vector<Task> tasks,
+                           CreateTasks(methods, fnames, memory_manager.get()));
 
       std::unique_ptr<ThreadPoolInternal> pool;
       std::vector<std::unique_ptr<ThreadPoolInternal>> inner_pools;
       InitThreads(tasks.size(), &pool, &inner_pools);
-
-      std::vector<PackedPixelFile> loaded_images = LoadImages(fnames, &*pool);
+      if (Args()->generations > 0) {
+        fprintf(stderr,
+                "Generation loss testing with %" PRIuS
+                " intermediate generations\n",
+                Args()->generations);
+      }
+      std::vector<PackedPixelFile> loaded_images =
+          LoadImages(fnames, pool->get());
 
       if (RunTasks(methods, extra_metrics_names, extra_metrics_commands, fnames,
-                   loaded_images, &*pool, inner_pools, &tasks) != 0) {
-        ret = EXIT_FAILURE;
+                   loaded_images, pool->get(), inner_pools, &tasks) != 0) {
+        ok = false;
         if (!Args()->silent_errors) {
           fprintf(stderr, "There were error(s) in the benchmark.\n");
         }
       }
     }
 
-    jxl::CacheAligned::PrintStats();
-    return ret;
+    PrintStats(memory_manager);
+    if (!ok) return JXL_FAILURE("RunTasks error");
+    return true;
   }
 
  private:
@@ -850,7 +924,7 @@ class Benchmark {
             " threads, %" PRIuS " inner threads\n",
             num_hw_threads, num_tasks, num_threads, num_inner);
 
-    pool->reset(new ThreadPoolInternal(num_threads));
+    *pool = jxl::make_unique<ThreadPoolInternal>(num_threads);
     // Main thread OR worker threads in pool each get a possibly empty nested
     // pool (helps use all available cores when #tasks < #threads)
     for (size_t i = 0; i < std::max<size_t>(num_threads, 1); ++i) {
@@ -890,7 +964,7 @@ class Benchmark {
         it = metrics.erase(it);
       } else {
         auto s = SplitString(*it, ':');
-        JXL_CHECK(s.size() == 2);
+        JPEGXL_TOOLS_CHECK(s.size() == 2);
         *it = s[1];
         ++it;
       }
@@ -901,7 +975,7 @@ class Benchmark {
   static StringVec SampleFromInput(const StringVec& fnames,
                                    const std::string& sample_tmp_dir,
                                    int num_samples, size_t size) {
-    JXL_CHECK(!sample_tmp_dir.empty());
+    JPEGXL_TOOLS_CHECK(!sample_tmp_dir.empty());
     fprintf(stderr, "Creating samples of %" PRIuS "x%" PRIuS " tiles...\n",
             size, size);
     StringVec fnames_out;
@@ -910,24 +984,27 @@ class Benchmark {
     size_t total_num_tiles = 0;
     for (const auto& fname : fnames) {
       Image3F img;
-      JXL_CHECK(ReadPNG(fname, &img));
-      JXL_CHECK(img.xsize() >= size);
-      JXL_CHECK(img.ysize() >= size);
+      JPEGXL_TOOLS_CHECK(ReadPNG(fname, &img));
+      JPEGXL_TOOLS_CHECK(img.xsize() >= size);
+      JPEGXL_TOOLS_CHECK(img.ysize() >= size);
       total_num_tiles += (img.xsize() - size + 1) * (img.ysize() - size + 1);
       offsets.push_back(total_num_tiles);
       images.emplace_back(std::move(img));
     }
-    JXL_CHECK(MakeDir(sample_tmp_dir));
+    JPEGXL_TOOLS_CHECK(MakeDir(sample_tmp_dir));
     Rng rng(0);
     for (int i = 0; i < num_samples; ++i) {
       int val = rng.UniformI(0, offsets.back());
       size_t idx = (std::lower_bound(offsets.begin(), offsets.end(), val) -
                     offsets.begin());
-      JXL_CHECK(idx < images.size());
+      JPEGXL_TOOLS_CHECK(idx < images.size());
       const Image3F& img = images[idx];
       int x0 = rng.UniformI(0, img.xsize() - size);
       int y0 = rng.UniformI(0, img.ysize() - size);
-      JXL_ASSIGN_OR_DIE(Image3F sample, Image3F::Create(size, size));
+      JXL_ASSIGN_OR_QUIT(
+          Image3F sample,
+          Image3F::Create(jpegxl::tools::NoMemoryManager(), size, size),
+          "Allocation failure.");
       for (size_t c = 0; c < 3; ++c) {
         for (size_t y = 0; y < size; ++y) {
           const float* JXL_RESTRICT row_in = img.PlaneRow(c, y0 + y);
@@ -939,7 +1016,7 @@ class Benchmark {
           StringPrintf("%s/%s.crop_%dx%d+%d+%d.png", sample_tmp_dir.c_str(),
                        FileBaseName(fnames[idx]).c_str(), size, size, x0, y0);
       ThreadPool* null_pool = nullptr;
-      JXL_CHECK(WriteImage(sample, null_pool, fn_output));
+      JPEGXL_TOOLS_CHECK(WriteImage(sample, null_pool, fn_output));
       fnames_out.push_back(fn_output);
     }
     fprintf(stderr, "Created %d sample tiles\n", num_samples);
@@ -948,9 +1025,9 @@ class Benchmark {
 
   static StringVec GetFilenames() {
     StringVec fnames;
-    JXL_CHECK(MatchFiles(Args()->input, &fnames));
+    JPEGXL_TOOLS_CHECK(MatchFiles(Args()->input, &fnames));
     if (fnames.empty()) {
-      JXL_ABORT("No input file matches pattern: '%s'", Args()->input.c_str());
+      JPEGXL_TOOLS_ABORT("No input file matches pattern");
     }
     if (Args()->print_details) {
       std::sort(fnames.begin(), fnames.end());
@@ -968,57 +1045,61 @@ class Benchmark {
                                                  ThreadPool* pool) {
     std::vector<PackedPixelFile> loaded_images;
     loaded_images.resize(fnames.size());
-    const auto process_image = [&](const uint32_t task, size_t /*thread*/) {
+    const auto process_image = [&](const uint32_t task,
+                                   size_t /*thread*/) -> Status {
       const size_t i = static_cast<size_t>(task);
-      Status ok = true;
+      Status ret = true;
 
       if (!Args()->decode_only) {
         std::vector<uint8_t> encoded;
-        ok = ReadFile(fnames[i], &encoded);
-        if (ok) {
-          ok = jxl::extras::DecodeBytes(Bytes(encoded), Args()->color_hints,
-                                        &loaded_images[i]);
+        ret = ReadFile(fnames[i], &encoded);
+        if (ret) {
+          ret = jxl::extras::DecodeBytes(Bytes(encoded), Args()->color_hints,
+                                         &loaded_images[i]);
         }
-        if (ok && loaded_images[i].icc.empty()) {
+        if (ret && loaded_images[i].icc.empty()) {
           // Add ICC profile if the image is not in sRGB, because not all codecs
           // can handle the color_encoding enum.
-          ok = CreateNonSRGBICCProfile(&loaded_images[i]);
+          ret = CreateNonSRGBICCProfile(&loaded_images[i]);
         }
-        if (ok && Args()->intensity_target != 0) {
+        if (ret && Args()->intensity_target != 0) {
           // TODO(szabadka) Respect Args()->intensity_target
         }
       }
-      if (!ok) {
+      if (!ret) {
         if (!Args()->silent_errors) {
           fprintf(stderr, "Failed to load image %s\n", fnames[i].c_str());
         }
-        return;
+        return JXL_FAILURE("Failed to load image");
       }
 
       if (!Args()->decode_only && Args()->override_bitdepth != 0) {
         // TODO(szabadla) Respect Args()->override_bitdepth
       }
+      return true;
     };
-    JXL_CHECK(jxl::RunOnPool(pool, 0, static_cast<uint32_t>(fnames.size()),
-                             ThreadPool::NoInit, process_image, "Load images"));
+    JPEGXL_TOOLS_CHECK(
+        jxl::RunOnPool(pool, 0, static_cast<uint32_t>(fnames.size()),
+                       ThreadPool::NoInit, process_image, "Load images"));
     return loaded_images;
   }
 
-  static std::vector<Task> CreateTasks(const StringVec& methods,
-                                       const StringVec& fnames) {
+  static StatusOr<std::vector<Task>> CreateTasks(
+      const StringVec& methods, const StringVec& fnames,
+      JxlMemoryManager* memory_manager) {
     std::vector<Task> tasks;
     tasks.reserve(methods.size() * fnames.size());
     for (size_t idx_image = 0; idx_image < fnames.size(); ++idx_image) {
       for (size_t idx_method = 0; idx_method < methods.size(); ++idx_method) {
         tasks.emplace_back();
         Task& t = tasks.back();
-        t.codec = CreateImageCodec(methods[idx_method]);
+        t.codec = CreateImageCodec(methods[idx_method], memory_manager);
         t.idx_image = idx_image;
         t.idx_method = idx_method;
         // t.stats is default-initialized.
       }
     }
-    JXL_ASSERT(tasks.size() == tasks.capacity());
+    JXL_ENSURE(tasks.size() == tasks.capacity());
     return tasks;
   }
 
@@ -1034,7 +1115,7 @@ class Benchmark {
       // Print CSV header
       printf(
           "method,image,error,size,pixels,enc_speed,dec_speed,"
-          "bpp,dist,psnr,p,bppp,qabpp");
+          "bpp,maxnorm,ssimulacra2,psnr,pnorm,bppp,qabpp");
       for (const std::string& s : extra_metrics_names) {
         printf(",%s", s.c_str());
       }
@@ -1042,25 +1123,29 @@ class Benchmark {
     }
 
     std::vector<uint64_t> errors_thread;
-    JXL_CHECK(jxl::RunOnPool(
-        pool, 0, tasks->size(),
-        [&](const size_t num_threads) {
-          // Reduce false sharing by only writing every 8th slot (64 bytes).
-          errors_thread.resize(8 * num_threads);
-          return true;
-        },
-        [&](const uint32_t i, const size_t thread) {
-          Task& t = (*tasks)[i];
-          const PackedPixelFile& image = loaded_images[t.idx_image];
-          t.image = &image;
-          std::vector<uint8_t> compressed;
-          DoCompress(fnames[t.idx_image], image, extra_metrics_commands,
-                     t.codec.get(), &*inner_pools[thread], &compressed,
-                     &t.stats);
-          printer.TaskDone(i, t);
-          errors_thread[8 * thread] += t.stats.total_errors;
-        },
-        "Benchmark tasks"));
+
+    const auto init = [&](const size_t num_threads) -> Status {
+      // Reduce false sharing by only writing every 8th slot (64 bytes).
+      errors_thread.resize(8 * num_threads);
+      return true;
+    };
+    const auto do_task = [&](const uint32_t i, const size_t thread) -> Status {
+      Task& t = (*tasks)[i];
+      const PackedPixelFile& image = loaded_images[t.idx_image];
+      t.image = &image;
+      std::vector<uint8_t> compressed;
+      if (!DoCompress(fnames[t.idx_image], image, extra_metrics_commands,
+                      t.codec.get(), inner_pools[thread]->get(), &compressed,
+                      &t.stats)) {
+        t.stats.total_errors++;
+      } else if (!printer.TaskDone(i, t)) {
+        t.stats.total_errors++;
+      }
+      errors_thread[8 * thread] += t.stats.total_errors;
+      return true;
+    };
+    JPEGXL_TOOLS_CHECK(jxl::RunOnPool(pool, 0, tasks->size(), init, do_task,
+                                      "Benchmark tasks"));
     if (Args()->show_progress) fprintf(stderr, "\n");
     return std::accumulate(errors_thread.begin(), errors_thread.end(),
                            static_cast<size_t>(0));
@@ -1071,22 +1156,22 @@ int BenchmarkMain(int argc, const char** argv) {
   fprintf(stderr, "benchmark_xl %s\n",
           jpegxl::tools::CodecConfigString(JxlDecoderVersion()).c_str());
 
-  JXL_CHECK(Args()->AddCommandLineOptions());
+  JPEGXL_TOOLS_CHECK(Args()->AddCommandLineOptions());
 
   if (!Args()->Parse(argc, argv)) {
     fprintf(stderr, "Use '%s -h' for more information\n", argv[0]);
-    return 1;
+    return EXIT_FAILURE;
   }
 
   if (Args()->cmdline.HelpFlagPassed()) {
     Args()->PrintHelp();
-    return 0;
+    return EXIT_SUCCESS;
   }
   if (!Args()->ValidateArgs()) {
     fprintf(stderr, "Use '%s -h' for more information\n", argv[0]);
-    return 1;
+    return EXIT_FAILURE;
   }
-  return Benchmark::Run();
+  return Benchmark::Run() ? EXIT_SUCCESS : EXIT_FAILURE;
 }
 
 }  // namespace
